/*
 * Simulator of microcontrollers (newcmd.cc)
 *
 * Copyright (C) 1999,99 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of microcontroller simulator: ucsim.

UCSIM is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

UCSIM is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with UCSIM; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

#include "ddconfig.h"

#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
# include <netinet/in.h>
# include <arpa/inet.h>
# include <netdb.h>
#endif
#include <sys/time.h>
#if FD_HEADER_OK
# include HEADER_FD
#endif
#include <unistd.h>
#include "i_string.h"

#include "globals.h"

#include "simcl.h"

#include "cmdcl.h"
#include "cmdutil.h"


/*
 * Command line
 */

cl_cmdline::cl_cmdline(char *acmd)
{
  cmd= strdup(acmd);
  params= new cl_list(2, 2);
  name= 0;
}

cl_cmdline::~cl_cmdline(void)
{
  if (cmd)
    free(cmd);
  if (name)
    free(name);
  delete params;
}

int
cl_cmdline::init(void)
{
  split();
  return(0);
}

int
cl_cmdline::split(void)
{
  char *start= cmd;
  int i;

  if (!cmd ||
      !*cmd)
    return(0);
  start+= strspn(start, " \t");
  if (start &&
      *start == '\n')
    {
      name= (char*)malloc(2);
      strcpy(name, "\n");
      return(0);
    }
  if (!*start)
    return(0);
  i= strcspn(start, " \t");
  if (i)
    {
      name= (char*)malloc(i+1);
      strncpy(name, start, i);
printf("found command name \"%s\"\n",name);
    }
  start+= i;
  while (*start &&
	 strchr(" \t\n\v\r,;:.", *start))
    start++;
  
}

int
cl_cmdline::repeat(void)
{
  return(name &&
	 *name == '\n');
}


/*
 * Command
 */

cl_cmd::cl_cmd(class cl_sim *asim,
	       char *aname,
	       int can_rep,
	       char *short_hlp,
	       char *long_hlp):
  cl_base()
{
  sim = asim;
  name= aname?strdup(aname):strdup("unknown");
  can_repeat= can_rep;
  short_help= short_hlp?strdup(short_hlp):NULL;
  long_help= long_hlp?strdup(long_hlp):NULL;
}

cl_cmd::~cl_cmd(void)
{
  if (name)
    free(name);
  if (short_help)
    free(short_help);
  if (long_help)
    free(long_help);
}

int
cl_cmd::work(class cl_cmdline *cmdline, class cl_console *console)
{
  fprintf(console->out,
	  "Command \"%s\" executed.\n", name);
  return(0);
}


/*
 * Set of commands
 */

cl_cmdset::cl_cmdset(class cl_sim *asim):
  cl_list(5, 5)
{
  sim= asim;
  last_command= 0;
}

class cl_cmd *
cl_cmdset::get_cmd(class cl_cmdline *cmdline)
{
  int i;

  if (cmdline->repeat())
    {
      if (last_command)
	return(last_command);
      else
	return(0);
    }
  for (i= 0; i < count; i++)
    {
      class cl_cmd *c= (class cl_cmd *)at(i);
      if (c->name &&
	  cmdline->name &&
	  strcmp(c->name, cmdline->name) == 0)
	{printf("GOT\n");return(c);}
    }
  return(0);
}


/*
 * Command console
 */

cl_console::cl_console(char *fin, char *fout, class cl_sim *asim):
  cl_base()
{
  FILE *f;

  last_command= NULL;
  sim= asim;
  in= stdin;
  if (fin)
    if (f= fopen(fin, "r+"), in= f?f:stdin, !f)
      fprintf(stderr, "Can't open `%s': %s\n", fin, strerror(errno));
  out= stdout;
  if (fout)
    if (f= fopen(fout, "w+"), out= f?f:stdout, !f)
      fprintf(stderr, "Can't open `%s': %s\n", fout, strerror(errno));
}

cl_console::cl_console(FILE *fin, FILE *fout, class cl_sim *asim):
  cl_base()
{
  last_command= NULL;
  sim= asim;
  in = fin;
  out= fout;
}

/*
 * use the port number supplied to connect to localhost for 
 * (by Sandeep)
 */

#ifdef SOCKET_AVAIL
static int
connect_to_port(int portnum)
{
  int sock= socket(AF_INET,SOCK_STREAM,0);
  struct sockaddr_in sin;

  sin.sin_family = AF_INET;
  sin.sin_port = htons(portnum);
  sin.sin_addr.s_addr = inet_addr("127.0.0.1");

  if (connect(sock, (struct sockaddr *)&sin, sizeof(sin))) {
    fprintf(stderr, "Connect to port %d: %s\n", portnum, strerror(errno));
    return -1;
  }
  return sock;
}

cl_console::cl_console(int portnumber, class cl_sim *asim)
{
  int sock= connect_to_port(portnumber);

  last_command= NULL;
  sim= asim;
  if (!(in= fdopen(sock, "r+")))
    fprintf(stderr, "cannot open port for input\n");
  if (!(out= fdopen(sock, "w+")))
    fprintf(stderr, "cannot open port for output\n");
  //fprintf(stderr, "init socket done\n");
}
#endif

int
cl_console::init(void)
{
  cl_base::init();
  welcome();
  prompt();
}

cl_console::~cl_console(void)
{
  if (in)
    fclose(in);
  if (out)
    {
      fprintf(out, "\n");
      fflush(out);
      fclose(out);
    }
#ifdef SOCKET_AVAIL
  /*  if (sock)
    {
      shutdown(sock, 2);
      close(sock);
      }*/
#endif
}

void
cl_console::welcome(void)
{
  fprintf(out,
	  "ucsim %s, Copyright (C) 1997 Daniel Drotos, Talker Bt.\n"
	  "ucsim comes with ABSOLUTELY NO WARRANTY; for details type "
	  "`show w'.\n"
	  "This is free software, and you are welcome to redistribute it\n"
	  "under certain conditions; type `show c' for details.\n",
	  VERSIONSTR);
  fflush(out);
}

void
cl_console::prompt(void)
{
  char *p;

  if (sim->arg_avail('P'))
    putc('\0', out);
  else
    fprintf(out, "%s", sim->arg_avail('P')?"\0":
	    ((p= sim->get_sarg(0, "prompt"))?p:"> "));
  fflush(out);
}

int
cl_console::match(int fdnum)
{
  if (in &&
      fileno(in) == fdnum)
    return(1);
  return(0);
}

int
cl_console::get_in_fd(void)
{
  return(in?fileno(in):-1);
}

char *
cl_console::read_line(void)
{
  char *s= NULL, *p;
  size_t n= 30;

#ifdef HAVE_GETLINE
  if (getline(&s, 0, in) < 0)
    return(0);
#else

# ifdef HAVE_GETDELIM
  s= (char *)malloc(n);
  if (getdelim(&s, &n, '\n', in) < 0)
    {
      free(s);
      return(0);
    }
# else

#  ifdef HAVE_FGETS
  s= (char *)malloc(300);
  if (fgets(s, 300, in) == NULL)
    {
      free(s);
      return(0);
    }
#  endif
# endif
#endif
  s[strlen(s)-1]= '\0';
  if (s[strlen(s)-1] == '\r')
    s[strlen(s)-1]= '\0';
  return(s);
}

int
cl_console::proc_input(void)
{
  int retval;

  if (feof(in))
    {
      fprintf(out, "End\n");
      return(1);
    }
  char *cmd= read_line();
  if (!cmd)
    return(1);
  fprintf(out, "got command: \"%s\"\n", cmd);
  retval= sim->do_cmd(cmd, this);
  prompt();
  free(cmd);
  return(retval);
}

int
cl_console::interpret(char *cmd)
{
  return(cmd &&
	 cmd[0] == 'q');
}


/*
 * This console listen on a socket and can accept connection requests
 */
#ifdef SOCKET_AVAIL

cl_listen_console::cl_listen_console(int serverport, class cl_sim *asim)
{
  last_command= NULL;
  sim= asim;
  if ((sock= make_server_socket(serverport)) >= 0)
    {
      if (listen(sock, 10) < 0)
	fprintf(stderr, "Listen on port %d: %s\n", strerror(errno));
    }
  in= out= 0;
}

int
cl_listen_console::match(int fdnum)
{
  return(sock == fdnum);
}

int
cl_listen_console::get_in_fd(void)
{
  return(sock);
}

int
cl_listen_console::proc_input(void)
{
  int newsock;
  int size;
  struct sockaddr_in sock_addr;

  size= sizeof(struct sockaddr); 
  newsock= accept(sock, (struct sockaddr*)&sock_addr, &size);
  if (newsock < 0)
    {
      perror("accept");
      return(0);
    }
  if (!(in= fdopen(newsock, "r+")))
    fprintf(stderr, "cannot open port for input\n");
  if (!(out= fdopen(newsock, "w+")))
    fprintf(stderr, "cannot open port for output\n");
  sim->cmd->add_console(sim->cmd->mk_console(in, out, sim));
  in= out= 0;
  return(0);
}

#endif /* SOCKET_AVAIL */


/*
 * Command interpreter
 */

cl_commander::cl_commander(class cl_sim *asim):
  cl_base()
{
  cons= new cl_list(1, 1); 
  sim= asim;
}

int
cl_commander::init(void)
{
  cl_base::init();
  if (!sim)
    return(1);
  if (sim->arg_avail('c'))
    {
      add_console(mk_console(sim->get_sarg('c', 0),
			     sim->get_sarg('c', 0), sim));
    }
  if (sim->arg_avail('Z'))
    {
      add_console(mk_console(sim->get_iarg(0, "Zport"), sim));
    }
#ifdef SOCKET_AVAIL
  if (sim->arg_avail('r'))
    {
      add_console(mk_console(sim->get_iarg('r', 0), sim));
    }
#endif
  if (cons->get_count() == 0)
    {
      add_console(mk_console(stdin, stdout, sim));
    }
}

cl_commander::~cl_commander(void)
{
  delete cons;
}

class cl_console *
cl_commander::mk_console(char *fin, char *fout, class cl_sim *asim)
{
  return(new cl_console(fin, fout, asim));
}

class cl_console *
cl_commander::mk_console(FILE *fin, FILE *fout, class cl_sim *asim)
{
  return(new cl_console(fin, fout, asim));
}

#ifdef SOCKET_AVAIL
class cl_console *
cl_commander::mk_console(int portnumber, class cl_sim *asim)
{
  return(new cl_listen_console(portnumber, asim));
}
#endif

void
cl_commander::add_console(class cl_console *console)
{
  console->init();
  cons->add(console);
  set_fd_set();
}

void
cl_commander::del_console(class cl_console *console)
{
  cons->disconn(console);
  set_fd_set();
}

void
cl_commander::set_fd_set(void)
{
  int i;
  
  FD_ZERO(&read_set);
  fd_num= 0;
  for (i= 0; i < cons->count; i++)
    {
      int fd;
      class cl_console *c= (class cl_console*)(cons->at(i));
      if ((fd= c->get_in_fd()) >= 0)
	{
	  if (fd > fd_num)
	    fd_num= fd;
	  FD_SET(fd, &read_set);
	}
    }
  fd_num++;
}

extern "C" vasprintf();
extern "C" vsnprintf();

static void
cmd_do_print(FILE *f, char *format, va_list ap)
{
#ifdef HAVE_VASPRINTF
  char *msg= NULL;
  vasprintf(&msg, format, ap);
  fprintf(f, "%s", msg);
  free(msg);
#else
#  ifdef HAVE_VSNPRINTF
  char msg[80*25];
  vsnprintf(msg, 80*25, format, ap);
  fprintf(f, "%s", msg);
#  else
#    ifdef HAVE_VPRINTF
  char msg[80*25];
  vsprintf(msg, format, ap); /* Dangerous */
  fprintf(f, "%s", msg);
#    else
#      ifdef HAVE_DOPRNT
  /* ??? */
  /*strcpy(msg, "Unimplemented grp_printf has called.\n");*/
#      else
  /*strcpy(msg, "grp_printf can not be implemented, upgrade your libc.\n");*/
#      endif
#    endif
#  endif
#endif
  fflush(f);
}

/*
 * Printing to all consoles
 */

int
cl_commander::all_printf(char *format, ...)
{
  va_list ap;
  int i;

  for (i= 0; i < cons->count; i++)
    {
      class cl_console *c= (class cl_console*)(cons->at(i));
      if (c->out)
	{
	  va_start(ap, format);
	  cmd_do_print(c->out, format, ap);
	  va_end(ap);
	}
    }
}

/*
 * Printing to actual_console
 */

int
cl_commander::printf(char *format, ...)
{
  va_list ap;
  int i;

  if (actual_console &&
      actual_console->out)
    {
      va_start(ap, format);
      cmd_do_print(actual_console->out, format, ap);
      va_end(ap);
    }
}

/*
 * Printing to consoles which have CONS_DEBUG flag set
 */

int
cl_commander::debug(char *format, ...)
{
  va_list ap;
  int i;

  for (i= 0; i < cons->count; i++)
    {
      class cl_console *c= (class cl_console*)(cons->at(i));
      if (c->out &&
	  c->flags & CONS_DEBUG)
	{
	  va_start(ap, format);
	  cmd_do_print(c->out, format, ap);
	  va_end(ap);
	}
    }
}

int
cl_commander::input_avail(void)
{
  struct timeval tv;
  int i;

  tv.tv_sec= tv.tv_usec= 0;
  active_set= read_set;
  i= select(fd_num/*FD_SETSIZE*/, &active_set, NULL, NULL, &tv);
  return(i);
}

int
cl_commander::wait_input(void)
{
  int i;

  active_set= read_set;
  i= select(fd_num/*FD_SETSIZE*/, &active_set, NULL, NULL, NULL);
  return(i);
}

int
cl_commander::proc_input(void)
{
  int i;

  for (i= 0; i < fd_num/*FD_SETSIZE*/; i++)
    if (FD_ISSET(i, &active_set))
      {
	class cl_console *c;
	int j;
	for (j= 0; j < cons->count; j++)
	  {
	    c= (class cl_console*)(cons->at(j));
	    if (c->match(i))
	      {
		actual_console= c;
		int retval= c->proc_input();
		if (retval)
		  {
		    del_console(c);
		    delete c;
		  }
		actual_console= 0;
		return(cons->count == 0);
	      }
	  }
      }
  return(0);
}


/* End of newcmd.cc */
